/******************************************************************************
 * $Id: tp3.c,v 1.2 2003/05/06 09:46:05 soldani Exp $
 *
 * BURNOTTE Christophe (2ELIN), SOLDANI Cyril (2ELIN), XHROUET Christophe (2EN)
 *
 * Structure des ordinateurs : Travail pratique 3
 *****************************************************************************/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "tp3.h"
#include "rand.h"

/* Prototype des fonctions auxiliaires */
void emitter(int input, int output, int timer);
void loss(int rate, int input, int output);
void receiver(int input, int output, int timer);
void timer(int input, int output);

/* Variable globale pour la gestion des signaux de fermeture */
int mustQuit = 0;

/* Gestionnaire de signal pour les signaux de terminaison et d'interruption. */
static void termination() {
  /* Modifie la variable globale mustQuit pour indiquer la fermeture */
  mustQuit = 1;
}

/* Gestionnaire de signal pour le signal d'alarme */
static void timeOut() {
}

/* Point d'entre du programme.
 * ENTRE
 * - argc : nombre d'arguments sur la ligne de commande (+ nom du programme)
 * - argv : tableau des arguments sur la ligne de commande
 * SORTIE
 * - EXIT_SUCCESS si tout s'est bien pass et EXIT_FAILURE sinon */
int main(int argc, char **argv) {
  /* Files de messages */
  int msgid[6];
  /* Taux d'erreur (en %) */
  int rate;
  /* PID */
  int pid;
  /* Structure pour le gestionnaire de signal */
  struct sigaction action;
  /* Autres variables */
  int i;
  char *ptr;

  /* Vrifie le nombre d'arguments */
  if (argc != 2) {
    puts("Usage: tp3 LOSS_RATE (in %)\n");
    return EXIT_FAILURE;
  }

  /* Lit le taux d'erreur */
  rate = (int) strtoul(argv[1], &ptr, 10);
  if ((ptr[0] != '\0') || (rate < 0) || (rate > 100)) {
    printf("%s is not a valid loss percentage\n", argv[1]);
    return EXIT_FAILURE;
  }
  
  /* Cration des files de message */
  for (i = 0; i < 6; i++)
    if ((msgid[i] = msgget(IPC_PRIVATE, 0600)) == -1) {
      /* Erreur */
      perror("Could not create message queue");
      for (;i > 0; i--)
	msgctl(msgid[i], IPC_RMID, NULL);
      return EXIT_FAILURE;
    }

  /* Cration des diffrents processus */
  for (i = 0; i < 5; i++) {
    if ((pid = fork()) == -1) {
      perror("Could not create process");
      kill(0, SIGKILL);
      return EXIT_FAILURE;
    }
    if (pid == 0)
      /* Fils, arrte la duplication */
      break;
  }

  if (pid == 0) {
  /* Attribution des comportements des fils */
    switch (i) {
      case 0:
	/* Bruiteur de message */
	loss(rate, msgid[MSG_NOISE], msgid[RECEIVER]);
	break;
      case 1:
	/* Rcepteur */
	receiver(msgid[RECEIVER], msgid[ACK_NOISE], msgid[ACK_TIMER]);
	break;
      case 2:
	/* Bruiteur d'acquis */
	loss(rate, msgid[ACK_NOISE], msgid[EMITTER]);
	break;
      case 3:
	/* Timer pour l'envoi de message */
	timer(msgid[MSG_TIMER], msgid[EMITTER]);
	break;
      case 4:
	/* Timer pour l'envoi d'acquis */
	timer(msgid[ACK_TIMER], msgid[RECEIVER]);
    }
    /* Quitte le fils */
    return 0;
  }

  /* Capte les signaux d'interruption et de terminaison */
  action.sa_handler = termination;
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  if (sigaction(SIGINT, &action, NULL) < 0) {
    /* Erreur */
    perror("Could not install SIGINT handler");
    return EXIT_FAILURE;
  }
  if (sigaction(SIGTERM, &action, NULL) < 0) {
    /* Erreur */
    perror("Could not install SIGTERM handler");
    return EXIT_FAILURE;
  }

  /* metteur (pre) */
  emitter(msgid[EMITTER], msgid[MSG_NOISE], msgid[MSG_TIMER]);

  /* Achve les zombies (by Lathander, vade retro satanas) */
  for (i = 0; i < 5; i++)
    pid = wait(NULL);

  /* Supprime les files de messages */
  for (i = 0; i < 6; i++)
    msgctl(msgid[i], IPC_RMID, NULL);

  /* Quitte le programme */
  return EXIT_SUCCESS;
}

/* metteur. Transmet des messages. Le protocole est le protocole du bit
 * altern.
 * ENTRE
 * - input : file d'entre
 * - output : file de sortie
 * - timer : file pour la gestion des time-out */
void emitter(int input, int output, int timer) {
  /* Structure de message */
  Message message;
  size_t messageSize = sizeof(message) - sizeof(long);
  /* Numro du message */
  int number = 0;
  /* Contenu du message (numro) */
  int content = 0;
  
  while (!mustQuit) {
    /* Envoie un message */
    message.mtype = MESSAGE;
    message.number = number;
    message.content = content;
    printf("* Sending message %d with number %d\n", message.content,
	message.number);
    fflush(stdout);
    msgsnd(output, &message, messageSize, 0);
    /* Prpare le timer */
    message.mtype = START;
    msgsnd(timer, &message, messageSize, 0);
    /* Attend l'acquis, le time-out ou le message de fermeture */
    msgrcv(input, &message, messageSize, 0, 0);
    if (message.mtype == END)
      /* Fin du programme */
      return;
    else if (message.mtype == ACK) {
      printf("  Received ack %d\n", message.number);
      fflush(stdout);
      /* Dsactive le timer */
      message.mtype = STOP;
      msgsnd(timer, &message, messageSize, 0);
      /* Aquis reu, est-ce le bon */
      if (message.number == number) {
	/* Oui, passe au message suivant */
	number = (number == 0) ? 1 : 0;
	content++;
	sleep(1);
      }
    }
    else {
      puts("  Sender time-out");
      fflush(stdout);
    }
  }

  /* On doit quitter le programme */
  /* Envoie le message de fermeture */
  message.mtype = END;
  msgsnd(output, &message, messageSize, 0);
  message.mtype = END;
  msgsnd(timer, &message, messageSize, 0);
  return;
}

/* Gnrateur alatoire de perte. Transmet les messages de la file d'entre 
 * la file de sortie en perdant alatoirement les messages au taux donn.
 * ENTRE
 * - rate : taux de perte en %
 * - input : file d'entre
 * - output : file de sortie */
void loss(int rate, int input, int output) {
  /* Structure de message */
  Message message;
  size_t messageSize = sizeof(message) - sizeof(long);
  /* Autres variables */
  int ret;
  
  /* Initialisation de la graine pour le gnrateur de nombres alatoires */
  setSeed(getpid());

  /* Boucle tant que le message END n'est pas reu */
  while (1) {
    /* Attend un message sur la file d'attente d'entre */
    do
      ret = msgrcv(input, &message, messageSize, 0, 0);
    while ((ret < 0) && (errno == EINTR));
    
    /* Vrifie que le message n'est pas END */
    if (message.mtype == END)
      break;

    /* Transmet le message sauf s'il est perdu */
    if ((1 + (int) (100.0 * getRand() / (CUSTOM_RAND_MAX + 1.0))) > rate)
      /* Envoie le message */
      msgsnd(output, &message, messageSize, 0);
    else {
      if (message.mtype == MESSAGE) {
	printf("  Lost message %d with number %d\n", message.content,
	    message.number);
	fflush(stdout);
      }
      else {
	printf("  Lost ack %d\n", message.number);
	fflush(stdout);
      }
    }
  }

  msgsnd(output, &message, messageSize, 0);
  return;
}

/* Rcepteur.
 * ENTRE
 * - input : file d'entre
 * - output : file de sortie
 * - timer : file pour la gestion des time-out */
void receiver(int input, int output, int timer) {
  /* Structure de message */
  Message message;
  size_t messageSize = sizeof(message) - sizeof(long);
  /* Numro attendu */
  int number = 0;
  /* Autres variables */
  int ret;
  
  /* Boucle tant que le message END n'est pas reu */
  while (1) {
    /* Attend un message sur la file d'attente d'entre */
    do
      ret = msgrcv(input, &message, messageSize, 0, 0);
    while ((ret < 0) && (errno == EINTR));
    
    /* Vrifie que le message n'est pas END */
    if (message.mtype == END)
      break;

    if (message.mtype == TIME_OUT) {
      /* Time-out */
      message.mtype = ACK;
      message.number = (number == 0) ? 1 : 0;
      printf("  Receiver time-out: sending ack %d\n", message.number);
      fflush(stdout);
      msgsnd(output, &message, messageSize, 0);
      /* Rsactive le compteur */
      message.mtype = START;
      msgsnd(timer, &message, messageSize, 0);
      continue;
    }
    
    if (message.number != number) {
      /* Ce n'est pas le message attendu, envoie un aquis ngatif */
      message.mtype = ACK;
      message.number = (number == 0) ? 1 : 0;
      printf("  Received wrong message %d: sending ack %d\n", message.content,
	  message.number);
      msgsnd(output, &message, messageSize, 0);
    }
    else {
      /* Message OK, affiche et envoie l'aquis */
      printf("* Received good message %d: sending ack %d\n", message.content,
	  message.number);
      fflush(stdout);
      message.mtype = ACK;
      message.number = number;
      number = (number == 0) ? 1 : 0;
      msgsnd(output, &message, messageSize, 0);
    }
    /* Rsactive le compteur */
    message.mtype = START;
    msgsnd(timer, &message, messageSize, 0);
  }

  message.mtype = END;
  msgsnd(output, &message, messageSize, 0);
  msgsnd(timer, &message, messageSize, 0);
  return;
}

/* Timer. Envoie un message sur la file de sortie quand le dlai est expir.
 * ENTRE
 * - input : file d'entre
 * - output : file de sortie */
void timer(int input, int output) {
  /* Structure pour le gestionnaire de signal */
  struct sigaction action;
  /* Structure pour les messages */
  Message message;
  int messageSize = sizeof(message) - sizeof(long);
  /* Autres variables */
  int ret;

  /* Capte les signaux d'alarme */
  action.sa_handler = timeOut;
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  if (sigaction(SIGALRM, &action, NULL) < 0) {
    /* Erreur */
    perror("Could not install SIGALRM handler");
    exit(EXIT_FAILURE);
  }

  while (1) {
    /* Attends un message */
    ret = msgrcv(input, &message, messageSize, 0, 0);
    if ((ret < 0) && (errno == EINTR)) {
      /* Alarme */
      message.mtype = TIME_OUT;
      msgsnd(output, &message, messageSize, 0);
    }
    else {
      if (message.mtype == END)
	break;
      else if (message.mtype == START)
	alarm(3);
      else
	alarm(0);
    }
  }

  return;
}
