/******************************************************************************
 * tp2.c
 *
 * Christophe BURNOTTE (2ELIN), Cyril SOLDANI (2ELIN),
 * Christophe XHROUET (2EN).
 *
 * Structure des ordinateurs : Travail pratique 2.
 *****************************************************************************/

#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "debug.h"

/* Variable globale utilise pour la terminaison du programme par l'envoi d'un
 * signal SIGINT ou SIGTERM */
int end = 0;

/* Comportement du consommateur.
 * Arguments :
 * - semid       : ensemble de smaphores ;
 * - buffer      : zone tampon ;
 * - position    : position courante dans la zone tampon ;
 * - has_to_read : nombre de processus qui ont lu la valeur courante ;
 * - next_sem    : smaphore d'attente suivant ;
 * - n           : nombre de consommateurs ;
 * - b           : nombre de place dans le tampon ;
 * - cid         : identifiant du consommateur.
 * Valeur de retour : aucune. */
void consummer(int semid, int *buffer, int *position, int *has_to_read,
    int *next_sem, int n, int b, int cid) {
  /* Numro du smaphore d'attente suivant */
  int semnum = 3;
  /* Variable pour l'utilisation des smaphores */
  struct sembuf sop;

  while (!end) {
    /* Alterne les deux smaphores de contrle s2 et s3 */
    semnum = (semnum == 2) ? 3 : 2;

    /* Attend d'tre rveill en faisant un wait sur le smaphore d'attente */
    sop.sem_num = semnum;
    sop.sem_op = -1;
    sop.sem_flg = 0;
    semop(semid, &sop, 1);

    /* Entre dans la section critique par un wait sur le smaphore s0 */
    sop.sem_num = 0;
    semop(semid, &sop, 1);

    if (buffer[*position] == 0) {
      /* On a lu un zro => sortie de la section critique et fermeture */
      sop.sem_num = 0;
      sop.sem_op = 1;
      semop(semid, &sop, 1);
      _exit(EXIT_SUCCESS);
    }

    /* Imprime la valeur lue */
    printf("p%d %d\n", cid, buffer[*position]);
    fflush(stdout);

    if (--*has_to_read == 0) {  
      /* Dernier consommateur  lire cette valeur */
      TRACE("(p%d) Je suis le dernier.\n", cid);
      if (*position == b - 1) {  
	/* Dernire case du tampon */
	TRACE("(p%d) Plus rien  lire, je rveille le boss.\n", cid);
	/* Remet l'index  zro */
	*position = 0;
	/* Remet le nombre de consommateurs qui doivent lire  n */
	*has_to_read = n;
	/* Met  jour le smaphore d'attente suivant */
	*next_sem = (semnum == 2) ? 3 : 2;
	/* veille le producteur en signalant le smaphore s1 */
	sop.sem_num = 1;
	sop.sem_op = 1;
	semop(semid, &sop, 1);
      }
      else {
	/* Passe  la case suivante */
	(*position)++;
	/* Remet le nombre de consommateurs qui doivent lire  n */
	*has_to_read = n;
	/* veille les autres consommateurs en signalant le smaphore d'attente
	 * suivant */
	sop.sem_num = (semnum == 2) ? 3 : 2;
	sop.sem_op = n;
	semop(semid, &sop, 1);
      }
    }

    /* Sort de la section critique par un signal sur le smaphore s0 */
    sop.sem_num = 0;
    sop.sem_op = 1;
    semop(semid, &sop, 1);
  }

  /* Interruption */
  _exit(EXIT_FAILURE);
}

/* Comportement du producteur.
 * Arguments :
 * - semid       : ensemble de smaphores ;
 * - next_sem    : smaphore d'attente suivant ;
 * - buffer      : zone tampon ;
 * - file        : fichier texte  lire ;
 * - n           : nombre de consommateurs ;
 * - b           : taille de la zone tampon.
 * Valeur de retour : aucune. */
void productor(int semid, int *next_sem, int *buffer, FILE *file, int n,
    int b) {
  /* Indique si on a lu un zro ou non */
  int has_read_zero = 0;
  /* Variable pour l'utilisation des smaphores */
  struct sembuf sop;
  /* Variables diverses */
  int i;
  char buf[256];
  
  while (!end) {
    /* Remplissage de la zone tampon */
    for (i = 0; i < b; i++) {
      /* Lecture dans le fichier */
      if (!fgets(buf, 255, file)) {
	/* Erreur d'E/S ou fin de fichier, considr comme un zro */
	buffer[i] = 0;
	has_read_zero = 1;
	break;
      }
      /* Conversion en entier (si la ligne ne contient pas de numro, elle est
       * considre comme un zro */
      buffer[i] = atoi(buf);
      /* Placement du nombre lu dans le tampon */
      if (buffer[i] == 0) {
	/* Zro lu, arrt */
	has_read_zero = 1;
	break;
      }
    }

    TRACE("Le producteur a fait son boulot, les consommateurs peuvent y "
	"aller.\n");

    /* Lache la meute des consommateurs en mettant s2 ou s3  n */
    sop.sem_num = *next_sem;
    sop.sem_op = n;
    sop.sem_flg = 0;
    semop(semid, &sop, 1);

    if (!has_read_zero) {
      /* Aucun zro n'a t lu, on attend patiemment son rveil par un wait sur
       * le smaphore s1 */
      sop.sem_num = 1;
      sop.sem_op = -1;
      semop(semid, &sop, 1);
    }
    else {
      /* Un zro a t lu, on quitte */
      break;
    }
  }
}

/* Gestionnaire de signal pour les signaux de terminaison et d'interruption. */
static void termination() {
  /* Modifie la variable end pour indiquer la terminaison du programme */
  end = 1;
}

/* Point d'entre du programme.
 * Arguments :
 * - argc : nombre d'arguments sur la ligne de commande + 1 ;
 * - argv : tableau contenant le nom du programme et les arguments sur la ligne
 *          de commande.
 * Valeur de retour :
 * - EXIT_SUCCESS si tout s'est bien pass (0) ;
 * _ EXIT_FAILURE si une erreur s'est produite (1). */
int main(int argc, char **argv) {
  /* Fichier texte contenant les numros */
  FILE *file;
  /* Nombre de consommateurs */
  int n;
  /* Nombre de places dans le buffer partag */
  int b;
  /* Ensemble de smaphores */
  int semid;			/* Identifiant de l'ensemble */
  union semun {
    int val;
    struct semid_ds *buf;
    unsigned short int *array;
    struct seminfo *__buf;
  } arg;
  unsigned short initv[] = {	/* Valeurs initiales des smaphores */
    1,				/* s0 protge la section critique */
    0,				/* s1 contrle le producteur */
    0,				/* s2, s3 contrlent les consommateurs en */
    0				/* alternance */
  };
  /* Zone de mmoire partage */
  int shmid;		/* Identifiant de la zone */
  int *position;	/* Position courante dans le tampon */
  int *has_to_read;	/* Nombre de consommateurs qui doivent encore lire la
			   valeur courante */
  int *next_sem;	/* Smaphore d'attente suivant */
  int *buffer;		/* Tampon partag  b places */
  /* Gestionnaire de signal */
  struct sigaction action;
  /* Variables diverses */
  int i;
  pid_t pid;

  /* Vrifie le nombre d'argument */
  if (argc != 4) {
    puts("\nUsage: tp2 FICHIER n b\n");
    return EXIT_FAILURE;
  }
  
  /* Lit le nombre de consommateurs */
  n = atoi(argv[2]);
  if (n <= 0) {
    /* Nombre de consommateurs incorrect */
    printf("%d n'est pas un nombre de consommateurs valide.\n", n);
    return EXIT_FAILURE;
  }

  /* Lit le nombre de places dans le buffer */
  b = atoi(argv[3]);
  if (b <= 0) {
    /* Nombre de places incorrect */
    printf("Vous avez dj vu un tampon de %d places ?!\n", n);
    return EXIT_FAILURE;
  }

  /* Ouvre le fichier */
  if (!(file = fopen(argv[1], "r"))) {
    /* Erreur */
    printf("Impossible d'ouvrir %s: %s.\n", argv[1], strerror(errno));
    return EXIT_FAILURE;
  }

  /* Capte les signaux d'interruption et de terminaison */
  action.sa_handler = termination;
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  if (sigaction(SIGINT, &action, NULL) < 0) {
    /* Erreur */
    perror("Could not install SIGINT handler");
    return EXIT_FAILURE;
  }
  if (sigaction(SIGTERM, &action, NULL) < 0) {
    /* Erreur */
    perror("Could not install SIGTERM handler");
    return EXIT_FAILURE;
  }

  /* Cration des smaphores */
  if ((semid = semget(IPC_PRIVATE, 4, 0660)) == -1) {
    perror("Impossible de crer les smaphores");
    return EXIT_FAILURE;
  }

  /* Initialisation des smaphores */
  arg.array = initv;
  if (semctl(semid, 4, SETALL, arg)) {
    perror("Pas moyen d'initialiser les smaphores");
    return EXIT_FAILURE;
  }

  /* Cration de la zone de mmoire partage */
  if ((shmid = shmget(IPC_PRIVATE, (b + 3) * sizeof(int), 0660)) == -1) {
    perror("Impossible de crer une zone de mmoire partage");
    return EXIT_FAILURE;
  }

  /* Attache la zone de mmoire partage au processus */
  position = shmat(shmid, NULL, 0);
  if (position == (int *) -1) {
    perror("Pas moyen d'attacher la zone de mmoire partage");
    return EXIT_FAILURE;
  }
  has_to_read = position + sizeof(int);
  *has_to_read = n;
  next_sem = has_to_read + sizeof(int);
  *next_sem = 2;
  buffer = next_sem + sizeof(int);
  
  /* Cration des consommateurs */
  for (i = 0; i < n; i++) {
    if ((pid = fork()) == -1) {
      /* Erreur */
      perror("Mitose impossible");
      kill(0, SIGKILL);
      return EXIT_FAILURE;
    }
    if (pid == 0)
      /* Fils */
      break;
  }

  if (pid == 0) {
    /* Consommateur */
    TRACE("Un nouveau consommateur est n : p%d avec le PID %d.\n", i + 1,
	getpid());
    consummer(semid, buffer, position, has_to_read, next_sem, n, b, i + 1);
    return EXIT_SUCCESS;
  }

  /* Producteur */
  TRACE("Papa s'est mis  la production.\n");
  productor(semid, next_sem, buffer, file, n, b);

  /* Achve les zombies (by Lathander, vade retro satanas) */
  for (i = 0; i < n; i++) {
    pid = wait(NULL);
    TRACE("Vade retro russi sur le zombie de PID %d.\n", pid);
  }

  /* Supprime l'ensemble de smaphores */
  if (semctl(semid, 0, IPC_RMID)) {
    perror("Les smaphores s'incrustent");
    return EXIT_FAILURE;
  }

  /* Supprime la zone de mmoire partage */
  if (shmctl(shmid, IPC_RMID, NULL)) {
    perror("La zone de mmoire partage refuse de mourir");
    return EXIT_FAILURE;
  }

  /* Ferme le fichier */
  if (fclose(file)) {
    printf("Pas moyen de fermer %s:%s.\n", argv[1], strerror(errno));
    return EXIT_FAILURE;
  }

  if (end)
    /* La fermeture du programme vient d'un signal d'interruption ou de
     * terminaison */
    puts("Interruption : les donnes affiches peuvent tre incorrectes.");

  /* Quitte le programme avec succs */
  return EXIT_SUCCESS;
}
